#! /usr/bin/python

"""//Web service to deposit credits from a desktop grid to an institution account//

- File: deposit.py
  
- Arguments - Using the HTTP GET method:
 - dg_id -- The desktop grid identifier, as an integer
 - institution_id -- The institution identifier, as an integer
 - credit -- The number of credits to transfer, as a positive integer
 - > If some arguments are missing, a web form will be displayed.
 
- Displays a Web page containing:
 - ``<div class="qos">True</div>`` if the deposit is correctly completed
 - ``<div class="qos">True</div>`` if the deposit is denied
   
- Notes: 
 - A deposit will be denied if the DG credit balance is lower than the credit to transfer. 
 - A database configuration file named "db.cfg" must be present in the directory.

"""

import sys
import os
sys.path.append(os.path.join(sys.path[0],".."))
os.chdir(sys.path[0])
from tools import db_func
from tools import cgi_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)



def deposit(conn,dg_id,institution_id,credit):
    """Function implementig the deposit operation

    Arguments:
    dg_id --- The desktop grid identifier, as an integer
    institution_id --- The institution identifier, as an integer
    credit --- The number of credits to transfer, as a positive integer

    The function returns True if the bill is accepted, or False otherwise
    
    """
    if not isinstance(conn,db_func.db_connection) or conn.open==0:
        log.error("The conn argument must be an open connection to the database !")
        return None     
    if not isinstance(credit,int) or credit<=0:
        log.error("The credit argument must be a positive integer !")
        return None
    if not isinstance(dg_id,int):
        log.error("The dg_id argument must be an integer !")
        return None
    if not isinstance(institution_id,int):
        log.error("The institution_id argument must be an integer !")
        return None                

    c=conn.cursor()

    log.info("Retrieving DG with identifier %d..." % dg_id)
    try:
        c.execute("SELECT credit_balance,credit_spent FROM DesktopGrid WHERE dg_id=%d" % dg_id)
    except db_func.DBError as e:
        log.error("Error during the SQL command execution !")
        log.error(e)
        return None         
    res = c.fetchall()
    
    if len(res)==0:
        log.error("No such DG associated to this identifier in the database")
        return None
    elif len(res)>1:
        log.error("Several DGs with identifier %d in the database ! 'dg_id' should be table DesktopGrid PRIMARY KEY" % dg_id)
        return None

    try:
        dg_credit_balance = int(res[0][0])
        if dg_credit_balance<0:
            raise ValueError("Negative credit_balance")
    except ValueError as e:
        log.error("'credit_balance' is not a positive or zero integer in the table DesktopGrid")  
        log.error(e)
        return None
    
    try:    
        dg_credit_spent = int(res[0][1])
        if dg_credit_spent<0:
            raise ValueError("Negative credit_spent")
    except ValueError as e:
        log.error("'credit_spent' is not a positive or zero integer in the table DesktopGrid")  
        log.error(e)
        return None
    
    log.info("DG found, credit balance was %d, credit spent was %d" \
            % (dg_credit_balance,dg_credit_spent))

    if dg_credit_balance<credit:
        log.error("This DG has not earned enough credits (only %d)" % dg_credit_balance)
        return False
    
    log.info("Retrieving Institution with identifier %d" % institution_id)
    try:
        c.execute("SELECT credit_balance,credit_won FROM Institution WHERE institution_id=%d" % institution_id)
    except db_func.DBError as e:
        log.error("Error during the SQL command execution !")
        log.error(e)
        return None 
    res = c.fetchall()
    
    if len(res)==0:
        log.error("No such institution associated to this identifier")
        return None
    elif len(res)>1:
        log.error("Several institutions with identifier %d in the database ! 'institution_id' should be table Institution PRIMARY KEY" % dg_id)
        return None

    try:
        institution_credit_balance = int(res[0][0])
        if institution_credit_balance<0:
            raise ValueError("Negative credit_balance")
    except ValueError as e:
        log.error("'credit_balance' is not a positive or zero integer in the table Institution")  
        log.error(e)
        return None        
    
    try:   
        institution_credit_won = int(res[0][1])
        if institution_credit_won<0:
            raise ValueError("Negative credit_won")
    except ValueError as e:
        log.error("'credit_won' is not a positive or zero integer in the table Institution")  
        log.error(e)
        return None           
     
    log.info("Institution found, credit balance was %d, credit won was %d" \
            % (institution_credit_balance,institution_credit_won))

    log.info("Updating DG credit spent and balance")
    try:
        c.execute("UPDATE DesktopGrid SET credit_spent=%d,credit_balance=%d WHERE dg_id=%d" % (dg_credit_spent+credit,dg_credit_balance-credit,dg_id))
    except db_func.DBError as e:
        log.error("Error during the SQL command execution !")
        log.error(e)
        return None     
    log.info("Done! New credit balance is %d (spent %d)" % (dg_credit_balance-credit,dg_credit_spent+credit))

    log.info("Updating institution account")
    try:
        c.execute("UPDATE Institution SET credit_balance=%d,credit_won=%d WHERE institution_id=%d" % \
            (institution_credit_balance+credit,institution_credit_won+credit,institution_id))
    except db_func.DBError as e:
        log.error("Error during the SQL command execution !")
        log.error(e)
        return None             
    log.info("Done! New credit balance is %d (won %d)" % (institution_credit_balance+credit,institution_credit_won+credit))

    return True


if __name__=="__main__":
    cgi_func.print_header()

    if (cgi_func.get_form_value('dg_id') is None) or  (cgi_func.get_form_value('institution_id') is None) \
	    or (cgi_func.get_form_value('credit') is None):
        cgi_func.print_form(['dg_id','institution_id','credit'])

    else:
        try:
            dg_id = int(cgi_func.get_form_value('dg_id'))
            institution_id = int(cgi_func.get_form_value('institution_id'))
            credit = int(cgi_func.get_form_value('credit'))                
        except ValueError as e:
            log.error("Arguments must be integers")
            cgi_func.print_form(['dg_id','institution_id','credit'])        
        else:       
            conn=db_func.connect(db_func.read_db_config("db.cfg"))
            cgi_func.print_result( deposit(conn,dg_id,institution_id,credit))
            db_func.disconnect(conn)
        
    cgi_func.print_footer()


