#!/usr/bin/python

"""//Cloud Worker implementation on G5K//

- > Contains some function to start and stop Cloud Worker instance hosted on G5K, not using the libcloud API

- File: g5k_cloudworker.py

- Notes: 
 - A file, named "g5k.cfg", with G5K credential (G5K_LOGIN and G5K_PASS), G5K site (G5K_SITE) and G5K instance information (G5_CW_IMAGE_ID) are needed to start the instances. 
 - A SSH tunnel linking the port 22 of the G5K_SITE frontend to localhost on port 22222 must exists (``ssh -L 22222:frontend.G5K_SITE.grid5000.fr:22 G5K_LOGIN@access.G5K_SITE.grid5000.fr``)
 
"""

import paramiko

import time
import threading
import urllib
import httplib
import base64
import string
import json

import sys
import os
sys.path.append(os.path.join(sys.path[0],".."))
os.chdir(sys.path[0])

from tools import conf_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)


class G5K_CW_Starter(threading.Thread):
    """//Thread that starts a cloud worker in G5K and make it executes a list of commands//

    - Arguments:
     - list_of_commands -- The list of commands to be executed by the cloud worker, as a list of stings
     
    - Attributes:
     - execution_is_success -- When the thread terminated, True if the cloud worker correctly started and the execution of the list of commands succeed, False otherwise
     - specific_data -- When the thread terminated and the execution succeed, may contain some specific data related to the batch. 
    
    """     
    
    def __init__(self,list_of_commands,wall_time=1):
        threading.Thread.__init__(self)
        if not isinstance(list_of_commands,list) or len(list_of_commands)==0:
            raise Exception("No command to execute")
        for s in list_of_commands:
            if not isinstance(s,str):
                raise Exception("Invalid list of commands (must be a list of string)")
        
        self.param = conf_func.read_config("g5k.cfg")
        if not ("G5K_LOGIN" in self.param and \
                "G5K_PASS" in self.param and \
                "G5K_CW_IMAGE_ID" in self.param and \
                "G5K_SITE" in self.param):
            raise Exception("Missing parameters in \"g5k.cfg\"")

        self.list_of_commands = list_of_commands
        self.wall_time=wall_time
        self.execution_is_success=False       
        self.specific_data=None


    def run(self):
        log.info("Connecting to G5K")
        
        base_url = "https://%s:%s@api.grid5000.fr/2.0/grid5000/sites/%s/jobs" % (self.param["G5K_LOGIN"],self.param["G5K_PASS"],self.param["G5K_SITE"])
        
        """
        log.info("Retrieving the list of G5K images")
        try:
            images = conn.list_images() 
        except Exception as e:
            log.error("Error when retrieving the list of images !")
            log.error(e)
            return None
        
        cw_img = None
        for i in images:
            if self.ec2_param["EC2_CW_IMAGE_ID"] == i.id:
                cw_img=i        

        if cw_img == None:
            log.error("Cannot find CloudWorker image in EC2")
            return         

        log.info("CloudWorker image found as %s" % str(cw_img))

        log.info("Reading through the list of EC2 instance types to find the type we want for the CW")
        sizes = conn.list_sizes() 
        cw_size = None
        for s in sizes:
            if self.ec2_param["EC2_CW_SIZE_ID"] == s.id:
                cw_size=s
        
        if cw_size == None:
            log.error("Cannot find CloudWorker size in EC2")
            return 

        log.info("CloudWorker instance type found as %s" % str(cw_size))
        """
        log.info("Starting the instance")
        
        urlo = urllib.URLopener()
        urlo.addheader('Accept','application/json')
        urlo.addheader('Content: application/json')
        data = "resources=nodes=1&resources=walltime=%d&command=sleep %d&name=Test&type=allow_classic_ssh" % (self.wall_time,self.wall_time*3600)
        r = urlo.open(base_url,data=data)
        uid = json.loads("".join(r.readlines()))["uid"]
        log.info("The instance uid is %d" % uid)
        
        #As the instance has no IP address yet, we must wait for the starting process to be completed
        #Periodacally read the state of G5K job until it is started
        log.info("Waiting for the instance startup to be completed...")
        node_address = None
        waited_time = 0
        while node_address==None and waited_time<600:
            time.sleep(5)
            waited_time += 5
            urlo = urllib.URLopener()
            urlo.addheader('Accept','application/json')
            r = urlo.open(base_url+"/%d" % uid)
            res = json.loads("".join(r.readlines()))
            if res["state"]=="running":
                node_address=res["assigned_nodes"][0]
        
        if node_address==None:
            log.error("Cannot retrieve the started instance of the CloudWorker")
            log.error("Destroying the instance previously created")
            #cw_node_launch.destroy()
            return

        log.info("The instance is started as %s" % node_address)

        self.specific_data = str(uid)

        log.info("Connecting with SSH")
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.WarningPolicy())
        tries=0
        while tries<5:
            try:
                ssh.connect("localhost",port=2222,username=self.param["G5K_LOGIN"],password=self.param["G5K_PASS"])
                break
            except Exception as e:
                log.warn("Failing to connect")
                log.warn(e)                
                tries += 1
                if tries<5:
                    log.warn("Trying again in 5 seconds")
                    time.sleep(5)
                else:
                    log.error("Too much tries, unable to connect")
                    log.error("Destroying the CloudWorker instance")
                    cw_node_started.destroy()
                    return
        log.info("Connected to G5K access")
       
        log.info("Executing commands")        
        for cmd in self.list_of_commands:
            sshcmd = "OAR_JOB_ID=%d oarsh %s '%s'" % (uid,node_address,cmd)
            log.info("Executing %s" % sshcmd)
            (stdin, stdout, stderr) =  ssh.exec_command(sshcmd)
            log.info("The output is \n%s\n%s" % ("".join(stdout.readlines()),"".join(stdout.readlines())) )
        
        log.info("All commands executed, closing connection")
        self.execution_is_success = True
        ssh.close()

class G5K_CW_Stopper(threading.Thread):
    """//Thread that stops a cloud worker in G5K//

    - Arguments:
     - uid -- The job uid in G5K coresponding to the intance
     
    - Attributes:
     - stop_is_success -- When the thread terminated, True if the cloud worker correctly stopped, False otherwise

    """     

    def __init__(self,instance_id):
        threading.Thread.__init__(self)
        if not isinstance(instance_id,str):
            raise Exception("The instance_id must be a string")
        
        self.param = conf_func.read_config("g5k.cfg")
        if not ("G5K_LOGIN" in self.param and \
                "G5K_PASS" in self.param and \
                "G5K_CW_IMAGE_ID" in self.param and \
                "G5K_SITE" in self.param):
            raise Exception("Missing parameters in \"g5k.cfg\"")

        self.instance_id = instance_id
        self.stop_is_success = False
        
    def run(self):
        log.info("Connecting to G5K")

        auth = 'Basic ' + string.strip(base64.encodestring(self.param["G5K_LOGIN"]+ ':' + self.param["G5K_PASS"]))
        host = "api.grid5000.fr"
        base_url = "/2.0/grid5000/sites/%s/jobs" % (self.param["G5K_SITE"])
        
        headers = {"Authorization": auth, "Content": "application/json", "Accept": "application/json"}                
        conn = httplib.HTTPSConnection(host)
        conn.request("DELETE", base_url+"/%s" % self.instance_id, headers=headers)
        response = conn.getresponse()
        conn.close()
                
        print response.status
        if response.status>299:
            log.error("Error when terminating the instance")
        else:
            log.info("Instance destroyed")
            self.stop_is_success=True



if __name__=="__main__":
    cw = G5K_CW_Starter(['ls','pwd','hostname','touch ok'])
    cw.start()
    cw.join()
    if cw.execution_is_success:
        print "Press enter to destroy the cloud worker"
        x=raw_input()
        cwstop = G5K_CW_Stopper(cw.specific_data)
        cwstop.start()
