#! /usr/bin/python

"""//Program that monitor Cloud Worker currently running//

- > This program manages Cloud Worker currently running, by billing the resources being spent, when stopping them if needed. It should be started periodically to perform monitoring operations. 

- File: monitor_cloud_workers.py

- Notes: 
 - A database configuration file named "db.cfg" must be present in the directory. 
 - A file named "creditsystem.cfg" containing a line "address=<url>", with <url> the url to access to the Credit System web server, must be present in the directory. 

- Algorithm:
``` 
For each of running cloud workers, this function checks 
  if the worker finished his job, 
      stop it
  else if there is still some credits in the batch order, 
      bill the resource spent
  else 
      stop the worker
  if the worker stopped and no other workers are working 
                                    on the worker's batch, 
      ask the Credit System to pay the order  
```

"""
import sys



import sys
import os
sys.path.append(os.path.join(sys.path[0],".."))
os.chdir(sys.path[0])
from tools import conf_func
from tools import web_func
from tools import db_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)

import scheduler_func

def monitor_cloud_workers(conn):
    """Function implementing the cloud worker monitoring

    Arguments:
    conn -- The scheduler's DB connection, as a MySQL.connections.Connection object

    """    
    if not isinstance(conn,db_func.db_connection) or conn.open==0:
        log.error("The conn argument must be an open connection to the database !")
        return None    
            
    c=conn.cursor()
   
    try:
        c.execute("""SELECT cloud_worker_id,cloud_tech,cloud_host,cloud_worker_data,batch_id,batch_data,dg_name,dg_type,dg_plugin_url
            FROM CloudWorkers NATURAL JOIN QoSBatches NATURAL JOIN DesktopGrid WHERE CloudWorkers.batch_id!='None'""")
    except db_func.DBError as e:
        log.error("Error during the SQL command execution !")
        log.error(e)
        return None                  
    res=c.fetchall()

    if len(res)==0:
        log.info("No cloud workers to check")
    else:
        for r in res:
            try:
                cw_id=int(r[0])
                cw_tech=r[1]
                cw_host=r[2]
                cw_data=r[3]
                batch_id=r[4]
                batch_data=r[5]
                dg_name=r[6]
                dg_type=r[7]
                dg_plugin_url=r[8]
            except ValueError as e:
                log.error("Error when retrieving result from sql query %s" % str(r))
                log.error(e)
                return None
            
            log.info("Checking cloud worker %d in %s@%s, computing batch %s on %s@%s" \
                    % (cw_id,cw_tech,cw_host,batch_id,dg_type,dg_name))

            stop_cw=False
            if scheduler_func.is_compute_finished(cw_id,cw_tech,cw_host,batch_id,dg_type,dg_name):                
                
                log.info("Cloud worker %d finnished to compute batch %s, stopping it..." % (cw_id,batch_id))
                stop_cw=True

            else:
                log.info("Cloud worker still have to run... Billing ressources spent")
                b=bill(batch_id,CREDIT_SPENT_EACH_TURN)
                
                if b==None:
                    log.error("No result from bill(%s), stopping this cloud worker as a precaution" % batch_id)
                    stop_cw=True               
                elif b==False:
                    log.info("No more credits available for batch %s, stopping cloud worker %d" % (batch_id,cw_id))
                    stop_cw=True
                else:
                    log.info("Done!")

            if stop_cw==True:

                log.info("Stopping CloudWorker %d" % cw_id)
                
                r = scheduler_func.stop_cloud_worker(cw_id,cw_tech,cw_host,cw_data)
                if r==False:
                    log.error("Error when stopping CloudWorker !!")
                    return None

                log.info("CloudWorker stopped, updating DB status")
                try:
                    c.execute("UPDATE CloudWorkers SET batch_id='None',cloud_worker_data='' WHERE cloud_worker_id=%d" % (cw_id))
                except db_func.DBError as e:
                    log.error("Error during the SQL command execution !")
                    log.error(e)
                    return None      
               
                log.info("Checking if there is other cloud workers working on this batch...")
                try:
                    c.execute("SELECT cloud_worker_id FROM CloudWorkers WHERE batch_id='%s'" % batch_id)
                except db_func.DBError as e:
                    log.error("Error during the SQL command execution !")
                    log.error(e)
                    return None                      
                res=c.fetchall()

                if len(res)>0:
                    log.info("Other cloud workers are still working on this batch, doing nothing more")
                else:
                    log.info("No more cloud workers are working on this batch, asking the credit system to pay the batch order")
                    p=pay(batch_id)               
                    if (p==None):
                        log.error("Error when calling pay() function")
                    else:
                        log.info("Done")

                    log.info("Updating the scheduler database")
                    try:
                        c.execute("DELETE FROM QoSBatches WHERE batch_id='%s'" % batch_id)
                    except db_func.DBError as e:
                        log.error("Error during the SQL command execution !")
                        log.error(e)
                        return None      
                
                    log.info("Unconfigure grid %s@%s for QoS on batch %s by cloud worker %d" % (dg_type,dg_name,batch_id,cw_id))
                    scheduler_func.unconfigure_QoS(batch_id,dg_type,dg_plugin_url,batch_data,cw_id)


            log.info("Cloud worker %d checked" % cw_id)


def get_credits(batch_id):
    """Remote call of Credit System's get_credits
    
    The function calls and retrieves the result of the get_credits function from the Credit System web server.
    It needs a file named "creditsystem.cfg" containing the line "address=<url>", with <url> the url to access to the QoS Info web server, to be present in the directory. 
    
    Arguments:
    batch_id -- The batch identifier, as a string
    
    The function returns the number of credits in the batch order, as an integer
    
    """  
    if not isinstance(batch_id,str):
        log.error("get_credits() batch_id argument must be a string")
        return None    
    
    params=conf_func.read_config("creditsystem.cfg")
    
    url="http://"+params['address']+"/get_credits.py?batch_id="+batch_id

    res=web_func.call(url)
                                                    
    if len(res)!=1:
        log.error("There should be 1 result from url %s" % url)
        return None
    try:
        r=int(res[0])
        if r<0:
            raise ValueError("Negative number of credit")
    except ValueError as e:
        log.error("get_credits result must be a positive integer")
        log.error(e)
        return None
    else:
        return r        


def pay(batch_id):
    """Remote call of Credit System's pay
    
    The function calls the pay function from the Credit System web server.
    It needs a file named "creditsystem.cfg" containing the line "address=<url>", with <url> the url to access to the QoS Info web server, to be present in the directory. 
    
    Arguments:
    batch_id -- The batch identifier, as a string
       
    """  
    if not isinstance(batch_id,str):
        log.error("pay() batch_id argument must be a string")
        return None    
            
    params=conf_func.read_config("creditsystem.cfg")
    
    url="http://"+params['address']+"/pay.py?batch_id="+batch_id

    res=web_func.call(url)
                                                    
    if len(res)!=1:
        log.error("There should be 1 result from url %s" % url)
        return None
    if res[0]!="True":
        log.error("Result from url %s should be True" % url)
        return None
    return True


def bill(batch_id,credit):
    """Remote call of Credit System's bill
    
    The function calls and retrieves the result of the pay function from the Credit System web server.
    It needs a file named "creditsystem.cfg" containing the line "address=<url>", with <url> the url to access to the QoS Info web server, to be present in the directory. 
    
    Arguments:
    batch_id -- The batch identifier, as a string
    credit -- The number of credit to bill (spent by cloud ressources)
       
    The function returns True if there is enough credits for this bill, False otherwise
           
    """  
    if not isinstance(batch_id,str):
        log.error("bill() batch_id argument must be a string")
        return None
    if not isinstance(credit,int) or credit<0:
        log.error("bill() credit argument must be a positive integer")
        return None
                    
    params=conf_func.read_config("creditsystem.cfg")
    
    url="http://"+params['address']+"/bill.py?batch_id="+batch_id+"&credit="+str(credit)

    res=web_func.call(url)
                                                    
    if len(res)!=1:
        log.error("Several or no results from url %s" % url)
        return None
    if not (res[0]=="True" or res[0]=="False"):
        log.error("Result from url %s should be True or False" % url)
        return None
    if res[0]=="True": 
        return True
    else:
        return False

if __name__=="__main__":
    #Credits cost each time a cloud worker is seen running by monitor_cloud_workers() 
    CREDIT_SPENT_EACH_TURN=1

    conn=db_func.connect(db_func.read_db_config("db.cfg"))
    monitor_cloud_workers(conn)
    db_func.disconnect(conn)



